SystemOrganization addCategory: #CryptographyRSFEC!

Object subclass: #RSFECDecoder
	instanceVariableNames: 'field'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!

!RSFECDecoder class methodsFor: 'instance creation' stamp: 'rww 6/4/2021 08:41'!
newDecoder

	RSFECGenericGFWithPlugin isPluginAvailable ifTrue: [ ^RSFECDecoderWithPlugin new ].
	^RSFECDecoder new! !

!RSFECDecoder class methodsFor: 'instance creation' stamp: 'rww 6/2/2021 12:09'!
newField: field

	^ self newDecoder
		initializeField: field;
		yourself.
! !

!RSFECDecoder methodsFor: 'action' stamp: 'rww 6/2/2021 13:31'!
decode: received twoS: twoS

	| eval noError poly syndromeCoefficients sigmaOmega syndrome errorLocations errorMagnitudes omega sigma position |
	poly := RSFECGenericGFPoly newField: field coefficients: received.
	syndromeCoefficients := ByteArray new: twoS withAll: 0.
	noError := true.
	0 to: (twoS - 1)
		do: [:index |
			eval := poly evaluateAt: (field exp: (index + field generatorBase)).
			syndromeCoefficients at: (syndromeCoefficients size - index) put: eval.
			(eval = 0)
				ifFalse: [noError := false]].
	noError ifTrue: [^self].
		
	syndrome := RSFECGenericGFPoly newField: field coefficients: syndromeCoefficients.
	sigmaOmega := self runEuclideanAlgorithmPoly: (field buildMonomialDegree: twoS coefficient: 1) poly: syndrome rDegrees: twoS.
	sigma := sigmaOmega at: 1.
	omega := sigmaOmega at: 2.
	errorLocations := self findErrorLocations: sigma.
	errorMagnitudes := self findErrorMagnitudes: omega errorLocations: errorLocations.

	1 to: errorLocations size
		do: [:index |
			position := received size - 1 - (field log: (errorLocations at: index)).
			(position < 0)
				ifTrue: [RSFECError signal: 'Bad error location'].
			received at: (position + 1) put: (field addOrSubtract: (received at: (position + 1))  by: (errorMagnitudes at: index))].
! !

!RSFECDecoder methodsFor: 'action' stamp: 'rww 6/2/2021 13:04'!
findErrorLocations: errorLocator
	"// This is a direct application of Chien's search"

	| numErrors e result index | 
	numErrors := errorLocator degree.
	(numErrors = 1)
		ifTrue: [^ ByteArray with: (errorLocator coefficient: 1)].

	result := Array new: numErrors withAll: 0.
	e := 0.
	index := 1.
	[(index <= field size) & (e < numErrors)]
		whileTrue: [
			((errorLocator evaluateAt: index) = 0)
				ifTrue: [
					result at: (e + 1) put: (field inverse: index).
					e := e + 1].
			index := index + 1].

	(e = numErrors)
		ifFalse: [RSFECError signal: 'Error locator degree does not match number of roots'].
	^ result! !

!RSFECDecoder methodsFor: 'action' stamp: 'rww 6/2/2021 13:31'!
findErrorMagnitudes: errorEvaluator errorLocations: errorLocations
	"// This is directly applying Forney's Formula"

	| denominator result s term termPlus1 xiInverse |
	s := errorLocations size.
	result := ByteArray new: s withAll: 0.
	1 to: s
		do: [:i |
			xiInverse := field inverse: (errorLocations at: i).
			denominator := 1.

			1 to: s
				do: [:j |
					(i ~= j)
						ifTrue: [
							term := field multiply: (errorLocations at: j) by: xiInverse.
							termPlus1 := ((term bitAnd: 16r1) = 0)
								ifTrue: [term bitOr: 1]
								ifFalse: [term - 1].
							denominator := field multiply: denominator by: termPlus1]].
	result at: i put: (field multiply: (errorEvaluator evaluateAt: xiInverse) by: (field inverse: denominator)).
	(field generatorBase = 0)
		ifFalse: [result at: i put: (field multiply: (result at: i) by: xiInverse)]].
	^ result! !

!RSFECDecoder methodsFor: 'initialize-release' stamp: 'rww 5/19/2021 13:59'!
initializeField: f

	field := f.
	
! !

!RSFECDecoder methodsFor: 'action' stamp: 'rww 6/2/2021 13:07'!
runEuclideanAlgorithmPoly:a poly: b rDegrees: rDegrees

	| aPoly bPoly degreeDiff denominatorLeadingTerm dltInverse inverse omega q r rLast rLastLast scale sigma sigmaTildeAtZero t tLast tLastLast |
	(a degree < b degree)
		ifTrue: [
			aPoly := b.
			bPoly := a]
		ifFalse: [
			aPoly := a.
			bPoly := b.].

	rLast := aPoly.
	r := bPoly.
	tLast := field zero.
	t := field one.

	[r degree >= (rDegrees // 2)]
		whileTrue: [
			"Transcript cr; show: 'Running EuclideanAlgorithm...r degree: ', r degree, ' rDegrees: ', rDegrees."
			rLastLast := rLast.
			tLastLast := tLast.
			rLast := r.
			tLast := t.
			rLast isZero
				ifTrue: [RSFECError signal: 'r_{i-1} was zero'].
			r := rLastLast.
			q := field zero.
			denominatorLeadingTerm := rLast coefficient: (rLast degree).
			dltInverse := field inverse: denominatorLeadingTerm.
			[r degree >= rLast degree & r isZero not]
				whileTrue: [
					degreeDiff := r degree - rLast degree.
					scale := field multiply: (r coefficient: (r degree)) by: dltInverse.
					q := q addOrSubtractPoly: (field buildMonomialDegree: degreeDiff coefficient: scale).
					r := r addOrSubtractPoly: (rLast multiplyByMonomialDegree: degreeDiff coefficient: scale )].


		t := (q multiplyPoly: tLast) addOrSubtractPoly: tLastLast.

		(r degree >= rLast degree)
			ifTrue: [RSFECError signal: 'Division algorithm failed to reduce polynomial?']].

	sigmaTildeAtZero := t coefficient: 0.
	(sigmaTildeAtZero = 0)
		ifTrue: [RSFECError signal: 'sigmaTilde(0) was zero'].
	inverse := field inverse: sigmaTildeAtZero.
	sigma := t multiplyScalar: inverse.
	omega := r multiplyScalar: inverse.
	^ Array with: sigma with: omega! !

Object subclass: #RSFECDecoderWithPlugin
	instanceVariableNames: 'field'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/16/2021 04:56'!
callPrimDecode: received twoS: twoS


	self 
		primDecode: received 
		twoS: twoS 
		generatorBase: field generatorBase 
		fieldSize: field size 
		oneOop: field one coefficients
		zeroOop: field zero coefficients.
! !

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/16/2021 04:04'!
callPrimRunEuclideanAlgorithmPoly: a poly: b rDegrees: rDegrees

	| result |
	result := self 
		primRunEuclideanAlgorithmPoly: a coefficients
		poly: b coefficients
		rDegrees: rDegrees 
		fieldSize: field size 
		one: field one coefficients 
		zero: field zero coefficients.
	^ result! !

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/2/2021 14:04'!
decode: received twoS: twoS

	| eval noError poly syndromeCoefficients sigmaOmega syndrome errorLocations errorMagnitudes omega sigma position |
	poly := RSFECGenericGFPoly newField: field coefficients: received.
	syndromeCoefficients := ByteArray new: twoS withAll: 0.
	noError := true.
	0 to: (twoS - 1)
		do: [:index |
			eval := poly evaluateAt: (field exp: (index + field generatorBase)).
			syndromeCoefficients at: (syndromeCoefficients size - index) put: eval.
			(eval = 0)
				ifFalse: [noError := false]].
	noError ifTrue: [^self].
		
	syndrome := RSFECGenericGFPoly newField: field coefficients: syndromeCoefficients.
	sigmaOmega := self runEuclideanAlgorithmPoly: (field buildMonomialDegree: twoS coefficient: 1) poly: syndrome rDegrees: twoS.
	sigma := sigmaOmega at: 1.
	omega := sigmaOmega at: 2.

	errorLocations := self findErrorLocations: sigma.
	errorMagnitudes := self findErrorMagnitudes: omega errorLocations: errorLocations.

	1 to: errorLocations size
		do: [:index |
			position := received size - 1 - (field log: (errorLocations at: index)).
			(position < 0)
				ifTrue: [RSFECError signal: 'Bad error location'].
			received at: (position + 1) put: (field addOrSubtract: (received at: (position + 1))  by: (errorMagnitudes at: index))].
! !

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/6/2021 09:10'!
findErrorLocations: errorLocator
	"// This is a direct application of Chien's search"

	| numErrors result | 
	numErrors := errorLocator degree.
	self etrace: 'RSFEC' msg: (String streamContents: [:stream |
		stream 
			nextPutAll: 'findErrorLocations: { ';
			crlf; tab; nextPutAll: 'errorLocator: ', errorLocator, ', numErrors = ', numErrors, '}']).
	(numErrors = 1)
		ifTrue: [^ ByteArray with: (errorLocator coefficient: 1)].

	self etrace: 'RSFEC' msg: (String streamContents: [:stream |
		stream 
			nextPutAll: 'findErrorLocations: {';
			crlf; tab; nextPutAll: 'errorLocator coefficients = ', errorLocator coefficients class, '}']).

	result := self 
		primFindErrorLocationsDegree: errorLocator degree 
		coefficients: errorLocator coefficients 
		fieldSize: field size.

	^ result! !

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/6/2021 09:15'!
findErrorMagnitudes: errorEvaluator errorLocations: errorLocations
	"// This is directly applying Forney's Formula"

	| result | 
	self etrace: 'RSFEC' msg: (String streamContents: [:stream |
		stream 
			nextPutAll: 'findErrorMagnitudes:errorLocations: { ';
			crlf; tab; nextPutAll: 'errorEvaluator: ', errorEvaluator, ', ';
			crlf; tab;
			nextPutAll: ' errorLocations = ', errorLocations class, ', ';
			crlf; tab; 
			nextPutAll: 'errorEvaluator coefficients = ', errorEvaluator coefficients class, '}']).

	result := self 
		primFindErrorMagnitudesLocations: errorLocations 
		coefficients: errorEvaluator coefficients 
		generatorBase: field generatorBase 
		fieldSize: field size.
		
	^ result! !

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/16/2021 10:18'!
inImageDecode: received twoS: twoS

	| eval noError poly syndromeCoefficients sigmaOmega syndrome errorLocations errorMagnitudes omega sigma position |
	poly := RSFECGenericGFPoly newField: field coefficients: received.
	syndromeCoefficients := ByteArray new: twoS withAll: 0.
	noError := true.
	0 to: (twoS - 1)
		do: [:index |
			eval := poly evaluateAt: (field exp: (index + field generatorBase)).
			syndromeCoefficients at: (syndromeCoefficients size - index) put: eval.
			(eval = 0)
				ifFalse: [noError := false]].
	noError ifTrue: [^self].
		
	syndrome := RSFECGenericGFPoly newField: field coefficients: syndromeCoefficients.
	sigmaOmega := self runEuclideanAlgorithmPoly: (field buildMonomialDegree: twoS coefficient: 1) poly: syndrome rDegrees: twoS.
	sigma := sigmaOmega at: 1.
	omega := sigmaOmega at: 2.

	errorLocations := self findErrorLocations: sigma.
	errorMagnitudes := self findErrorMagnitudes: omega errorLocations: errorLocations.

	1 to: errorLocations size
		do: [:index |
			position := received size - 1 - (field log: (errorLocations at: index)).
			(position < 0)
				ifTrue: [RSFECError signal: 'Bad error location'].
			received at: (position + 1) put: (field addOrSubtract: (received at: (position + 1))  by: (errorMagnitudes at: index))].
! !

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/16/2021 10:18'!
inImageRunEuclideanAlgorithmPoly:a poly: b rDegrees: rDegrees

	| aPoly bPoly degreeDiff denominatorLeadingTerm dltInverse inverse omega q r rLast rLastLast scale sigma sigmaTildeAtZero t tLast tLastLast |
	(a degree < b degree)
		ifTrue: [
			aPoly := b.
			bPoly := a]
		ifFalse: [
			aPoly := a.
			bPoly := b.].

	rLast := aPoly.
	r := bPoly.
	tLast := field zero.
	t := field one.

	[r degree >= (rDegrees // 2)]
		whileTrue: [
			"Transcript cr; show: 'Running EuclideanAlgorithm...r degree: ', r degree, ' rDegrees: ', rDegrees."
			rLastLast := rLast.
			tLastLast := tLast.
			rLast := r.
			tLast := t.
			rLast isZero
				ifTrue: [RSFECError signal: 'r_{i-1} was zero'].
			r := rLastLast.
			q := field zero.
			denominatorLeadingTerm := rLast coefficient: (rLast degree).
			dltInverse := field inverse: denominatorLeadingTerm.
			[r degree >= rLast degree & r isZero not]
				whileTrue: [
					degreeDiff := r degree - rLast degree.
					scale := field multiply: (r coefficient: (r degree)) by: dltInverse.
					q := q addOrSubtractPoly: (field buildMonomialDegree: degreeDiff coefficient: scale).
					r := r addOrSubtractPoly: (rLast multiplyByMonomialDegree: degreeDiff coefficient: scale )].

		t := (q multiplyPoly: tLast) addOrSubtractPoly: tLastLast.

		(r degree >= rLast degree)
			ifTrue: [RSFECError signal: 'Division algorithm failed to reduce polynomial?']].

	sigmaTildeAtZero := t coefficient: 0.
	(sigmaTildeAtZero = 0)
		ifTrue: [^ RSFECError signal: 'sigmaTilde(0) was zero'].
	inverse := field inverse: sigmaTildeAtZero.
	sigma := t multiplyScalar: inverse.
	omega := r multiplyScalar: inverse.
	^ Array with: sigma with: omega! !

!RSFECDecoderWithPlugin methodsFor: 'initialize-release' stamp: 'rww 6/4/2021 09:27'!
initializeField: f

	field := f.
	TraceMonitor 
		globalMonitor monitor: self;
		disableDomain: 'RSFEC'.! !

!RSFECDecoderWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 20:23'!
primDecode: decoded twoS: twoS generatorBase: generatorBase fieldSize: fieldSize oneOop: oneOop zeroOop: zeroOop

	<primitive: 'primitiveFECDecodeOopTwoSGeneratorBaseFieldSizeOneOopZeroOop' module: 'RSPlugin'>
	^ self primitiveFailed

! !

!RSFECDecoderWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 14:02'!
primFindErrorLocationsDegree: degree coefficients: coefficients fieldSize: fieldSize

	<primitive: 'primitiveFECFindErrorLocationsDegreeCoefficientsOopFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed

! !

!RSFECDecoderWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 14:03'!
primFindErrorMagnitudesLocations: errorLocations coefficients: coefficients generatorBase: generatorBase fieldSize: fieldSize

	<primitive: 'primitiveFECFindErrorMagnitudesLocationsOopCoefficientsOopGeneratorBaseFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed
! !

!RSFECDecoderWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 14:03'!
primRunEuclideanAlgorithmPoly:a poly: b rDegrees: rDegrees fieldSize: fieldSize one: one zero: zero

	<primitive: 'primitiveFECRunEuclideanAlgorithmPolyAOopPolyBOoprDegreesFieldSizeOneOopZeroOop' module: 'RSPlugin'>
	^ self primitiveFailed
! !

!RSFECDecoderWithPlugin methodsFor: 'action' stamp: 'rww 6/16/2021 10:19'!
runEuclideanAlgorithmPoly:a poly: b rDegrees: rDegrees

	| aPoly bPoly degreeDiff denominatorLeadingTerm dltInverse inverse omega q r rLast rLastLast scale sigma sigmaTildeAtZero t tLast tLastLast |
	(a degree < b degree)
		ifTrue: [
			aPoly := b.
			bPoly := a]
		ifFalse: [
			aPoly := a.
			bPoly := b.].

	rLast := aPoly.
	r := bPoly.
	tLast := field zero.
	t := field one.

	[r degree >= (rDegrees // 2)]
		whileTrue: [
			"Transcript cr; show: 'Running EuclideanAlgorithm...r degree: ', r degree, ' rDegrees: ', rDegrees."
			rLastLast := rLast.
			tLastLast := tLast.
			rLast := r.
			tLast := t.
			rLast isZero
				ifTrue: [RSFECError signal: 'r_{i-1} was zero'].
			r := rLastLast.
			q := field zero.
			denominatorLeadingTerm := rLast coefficient: (rLast degree).
			dltInverse := field inverse: denominatorLeadingTerm.
			[r degree >= rLast degree & r isZero not]
				whileTrue: [
					degreeDiff := r degree - rLast degree.
					scale := field multiply: (r coefficient: (r degree)) by: dltInverse.
					q := q addOrSubtractPoly: (field buildMonomialDegree: degreeDiff coefficient: scale).
					r := r addOrSubtractPoly: (rLast multiplyByMonomialDegree: degreeDiff coefficient: scale )].

		t := (q multiplyPoly: tLast) addOrSubtractPoly: tLastLast.

		(r degree >= rLast degree)
			ifTrue: [RSFECError signal: 'Division algorithm failed to reduce polynomial?']].

	sigmaTildeAtZero := t coefficient: 0.
	(sigmaTildeAtZero = 0)
		ifTrue: [^ RSFECError signal: 'sigmaTilde(0) was zero'].
	inverse := field inverse: sigmaTildeAtZero.
	sigma := t multiplyScalar: inverse.
	omega := r multiplyScalar: inverse.
	^ Array with: sigma with: omega! !

Object subclass: #RSFECEncoder
	instanceVariableNames: 'field cachedGenerators'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!

!RSFECEncoder class methodsFor: 'instance creation' stamp: 'rww 5/19/2021 13:57'!
newField: field

	^ self new
		initializeField: field;
		yourself! !

!RSFECEncoder methodsFor: 'actions' stamp: 'rww 6/2/2021 13:06'!
buildGenerator: degree

	| lastGenerator nextGenerator cacheGeneratorsOC |
	cacheGeneratorsOC := cachedGenerators asOrderedCollection.
	(degree >= cacheGeneratorsOC size)
		ifTrue: [
			lastGenerator := cacheGeneratorsOC at: (cacheGeneratorsOC size).
			cacheGeneratorsOC size to: degree
				do: [:index |
						nextGenerator := lastGenerator multiplyPoly: (RSFECGenericGFPoly 
							newField: field 
							coefficients: (ByteArray with: 1 with: (field exp: (index - 1 + field generatorBase)))).
						cacheGeneratorsOC add: nextGenerator.
						lastGenerator := nextGenerator]].
	cachedGenerators := cacheGeneratorsOC asArray.
	^ cachedGenerators at: (degree + 1).! !

!RSFECEncoder methodsFor: 'actions' stamp: 'rww 6/14/2021 08:55'!
encode: toEncode ecSize: ecSize

	| coefficients dataBytes generator info infoCoefficients numZeroCoefficients remainder |
	(ecSize = 0)
		ifTrue: [RSFECError signal: 'No error correction bytes'].
	dataBytes := toEncode size - ecSize.
	(dataBytes <= 0)
		ifTrue: [RSFECError signal: 'No data bytes provided'].

	generator := self buildGenerator: ecSize.
	infoCoefficients := ByteArray new: dataBytes withAll: 0.
	infoCoefficients replaceFrom: 1 to: dataBytes with: toEncode startingAt: 1.
	info := RSFECGenericGFPoly newField: field coefficients: infoCoefficients.
	info := info multiplyByMonomialDegree: ecSize coefficient: 1.
	remainder := (info dividePoly: generator) at: 2.
	coefficients := remainder coefficients.
	numZeroCoefficients := ecSize - coefficients size.
	1 to: numZeroCoefficients
		do: [:index |
			toEncode at: dataBytes + index put: 0.].
	
	toEncode replaceFrom: (dataBytes + numZeroCoefficients + 1) to: toEncode size with: coefficients startingAt: 1.! !

!RSFECEncoder methodsFor: 'initialize-release' stamp: 'rww 5/26/2021 14:35'!
initializeField: f

	field := f.
	cachedGenerators := OrderedCollection new.
	cachedGenerators add: (RSFECGenericGFPoly newField: field coefficients: #(1)).
	cachedGenerators := cachedGenerators asArray.
! !

Object subclass: #RSFECGenericGF
	instanceVariableNames: 'expTable logTable zero one size primitive generatorBase'
	classVariableNames: 'AZTEC_DATA_10 AZTEC_DATA_12 AZTEC_DATA_6 AZTEC_DATA_8 AZTEC_PARAM DATA_MATRIX_FIELD_256 MAXICODE_FIELD_64 QR_CODE_FIELD_256'
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/26/2021 12:12'!
aztecData10

	^ AZTEC_DATA_10! !

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/26/2021 12:12'!
aztecData12

	^ AZTEC_DATA_12! !

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/26/2021 12:11'!
aztecData6

	^ AZTEC_DATA_6! !

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/26/2021 12:12'!
aztecData8

	^ AZTEC_DATA_8! !

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/25/2021 15:32'!
aztecParam

	^ AZTEC_PARAM! !

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/20/2021 15:23'!
dataMatrixField256

	^ DATA_MATRIX_FIELD_256! !

!RSFECGenericGF class methodsFor: 'class initialization' stamp: 'rww 4/17/2021 14:21'!
initialize

	super initialize.
	Smalltalk addToStartUpList: self.
	self startUp: true.! !

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/26/2021 12:10'!
maxicodeField64

	^ MAXICODE_FIELD_64! !

!RSFECGenericGF class methodsFor: 'instance creation' stamp: 'rww 5/20/2021 05:11'!
newGenericGF

	RSFECGenericGFWithPlugin isPluginAvailable ifTrue: [ ^RSFECGenericGFWithPlugin new ].
	^RSFECGenericGF new! !

!RSFECGenericGF class methodsFor: 'instance creation' stamp: 'rww 5/27/2021 08:47'!
newPrimitive: prim size: size generatorBase: b

	"RSFECGenericGF newPrimitive: 16r012D size: 256 generatorBase: 1"
	^ self newGenericGF
		initializePrimitive: prim size: size generatorBase: b;
		yourself! !

!RSFECGenericGF class methodsFor: 'accessing' stamp: 'rww 5/25/2021 15:34'!
qrCodeField256

	^ QR_CODE_FIELD_256! !

!RSFECGenericGF class methodsFor: 'class initialization' stamp: 'rww 6/14/2021 09:10'!
startUp: resuming
	"self startUp: false"
	
	AZTEC_DATA_12  := RSFECGenericGF newPrimitive: 16r1069 size: 4096 generatorBase: 1. "x^12 + x^6 + x^5 + x^3 + 1"
	AZTEC_DATA_10 := RSFECGenericGF newPrimitive: 16r409 size: 1024 generatorBase: 1. "x^10 + x^3 + 1"
	AZTEC_DATA_6 := RSFECGenericGF newPrimitive: 16r43 size: 64 generatorBase: 1. "x^6 + x + 1"
	AZTEC_PARAM := RSFECGenericGF newPrimitive: 16r13 size: 16 generatorBase: 1. "x^4 + x + 1"
	QR_CODE_FIELD_256 := RSFECGenericGF newPrimitive: 16r011D size: 256 generatorBase: 0. "x^8 + x^4 + x^3 + x^2 + 1"
	DATA_MATRIX_FIELD_256 := RSFECGenericGF newPrimitive: 16r012D size: 256 generatorBase: 1. "x^8 + x^5 + x^3 + x^2 + 1"
	AZTEC_DATA_8 := DATA_MATRIX_FIELD_256.
	MAXICODE_FIELD_64 := AZTEC_DATA_6.
! !

!RSFECGenericGF methodsFor: 'arithmetic' stamp: 'rww 5/30/2021 14:42'!
addOrSubtract: a by: b

	^ (self maskValue: a) bitXor: (self maskValue: b).
! !

!RSFECGenericGF methodsFor: 'building' stamp: 'rww 6/1/2021 13:35'!
buildMonomialDegree: degree coefficient: coefficient

	| coefficients |
	(degree < 0)
		ifTrue: [RSFECError signal: 'negative degree'].
	(coefficient = zero)
		ifTrue: [^ zero].
	coefficients := ByteArray new: (degree + 1) withAll: 0.
	coefficients at: 1 put: coefficient.
	^ RSFECGenericGFPoly newField: self coefficients: coefficients.! !

!RSFECGenericGF methodsFor: 'arithmetic' stamp: 'rww 5/26/2021 17:12'!
exp: a

	^expTable at: (self normalizeIndex: a).! !

!RSFECGenericGF methodsFor: 'accessing' stamp: 'rww 5/17/2021 16:43'!
generatorBase

	^ generatorBase! !

!RSFECGenericGF methodsFor: 'initialize-release' stamp: 'rww 6/4/2021 10:02'!
initializeExpTable

    "expTable = new int[size];
    int x = 1;
    for (int i = 0; i < size; i++) {
      expTable[i] = x;
      x *= 2; // we're assuming the generator alpha is 2
      if (x >= size) {
        x ^= primitive;
        x &= size - 1;
      }
    }"

	| x |
	expTable := Array new: size withAll: 0.

	x := 1.
	(0 to: (size - 1))
		do: [:index |
			expTable at: (index + 1) put: x.
			x := x * 2.
			(x >= size)
				ifTrue: [x := ((x bitXor: primitive) bitAnd: (size - 1)) bitAnd: 16rFF]].

	expTable := expTable.

	"expTable := #(1 2 4 8 16 32 64 128 29 58 116 232 205 135 19 38 76 152 45 90 180 117 234 201 143 3 6 12 24 48 96 192 157 39 78 156 37 74 148 53 106 212 181 119 238 193 159 35 70 140 5 10 20 40 80 160 93 186 105 210 185 111 222 161 95 190 97 194 153 47 94 188 101 202 137 15 30 60 120 240 253 231 211 187 107 214 177 127 254 225 223 163 91 182 113 226 217 175 67 134 17 34 68 136 13 26 52 104 208 189 103 206 129 31 62 124 248 237 199 147 59 118 236 197 151 51 102 204 133 23 46 92 184 109 218 169 79 158 33 66 132 21 42 84 168 77 154 41 82 164 85 170 73 146 57 114 228 213 183 115 230 209 191 99 198 145 63 126 252 229 215 179 123 246 241 255 227 219 171 75 150 49 98 196 149 55 110 220 165 87 174 65 130 25 50 100 200 141 7 14 28 56 112 224 221 167 83 166 81 162 89 178 121 242 249 239 195 155 43 86 172 69 138 9 18 36 72 144 61 122 244 245 247 243 251 235 203 139 11 22 44 88 176 125 250 233 207 131 27 54 108 216 173 71 142 1 2 4 8 16 32 64 128 29 58 116 232 205 135 19 38 76 152 45 90 180 117 234 201 143 3 6 12 24 48 96 192 157 39 78 156 37 74 148 53 106 212 181 119 238 193 159 35 70 140 5 10 20 40 80 160 93 186 105 210 185 111 222 161 95 190 97 194 153 47 94 188 101 202 137 15 30 60 120 240 253 231 211 187 107 214 177 127 254 225 223 163 91 182 113 226 217 175 67 134 17 34 68 136 13 26 52 104 208 189 103 206 129 31 62 124 248 237 199 147 59 118 236 197 151 51 102 204 133 23 46 92 184 109 218 169 79 158 33 66 132 21 42 84 168 77 154 41 82 164 85 170 73 146 57 114 228 213 183 115 230 209 191 99 198 145 63 126 252 229 215 179 123 246 241 255 227 219 171 75 150 49 98 196 149 55 110 220 165 87 174 65 130 25 50 100 200 141 7 14 28 56 112 224 221 167 83 166 81 162 89 178 121 242 249 239 195 155 43 86 172 69 138 9 18 36 72 144 61 122 244 245 247 243 251 235 203 139 11 22 44 88 176 125 250 233 207 131 27 54 108 216 173 71 142)"! !

!RSFECGenericGF methodsFor: 'initialize-release' stamp: 'rww 6/4/2021 10:03'!
initializeLogTable

	logTable := Array new: size withAll: 0.
	0 to: (size - 2)
		do: [:index |
			logTable at: ((expTable at: (index + 1)) + 1) put: index].
	logTable := logTable.

	"logTable := #(
            255 0    1   25    2   50   26  198
            3  223   51  238   27  104  199   75
            4  100  224   14   52  141  239  129
            28  193  105  248  200    8   76  113
            5  138  101   47  225   36   15   33
            53  147  142  218  240   18  130   69
            29  181  194  125  106   39  249  185
            201  154    9  120   77  228  114  166
            6  191  139   98  102  221   48  253
            226  152   37  179   16  145   34  136
            54  208  148  206  143  150  219  189
            241  210   19   92  131   56   70   64
            30   66  182  163  195   72  126  110
            107   58   40   84  250  133  186   61
            202   94  155  159   10   21  121   43
            78  212  229  172  115  243  167   87
            7  112  192  247  140  128   99   13
            103   74  222  237   49  197  254   24
            227  165  153  119   38  184  180  124
            17   68  146  217   35   32  137   46
            55   63  209   91  149  188  207  205
            144  135  151  178  220  252  190   97
            242   86  211  171   20   42   93  158
            132   60   57   83   71  109   65  162
            31   45   67  216  183  123  164  118
            196   23   73  236  127   12  111  246
            108  161   59   82   41  157   85  170
            251   96  134  177  187  204   62   90
            203   89   95  176  156  169  160   81
            11  245   22  235  122  117   44  215
            79  174  213  233  230  231  173  232
            116  214  244  234  168   80   88  175)."! !

!RSFECGenericGF methodsFor: 'initialize-release' stamp: 'rww 6/1/2021 11:40'!
initializePrimitive: prim size: mySize generatorBase: b

	primitive := prim.
	size := mySize.
	generatorBase := b.

	self initializeExpTable.
	self initializeLogTable.

	zero := RSFECGenericGFPoly newField: self coefficients: #( 0) asByteArray.
	one := RSFECGenericGFPoly newField: self coefficients: #( 1) asByteArray.
! !

!RSFECGenericGF methodsFor: 'arithmetic' stamp: 'rww 5/26/2021 17:14'!
inverse: a

	(a = 0)
		ifTrue: [RSFECError signal: 'the argument is zero'].
	^ self exp: (size - (self log: a) - 1).
! !

!RSFECGenericGF methodsFor: 'arithmetic' stamp: 'rww 6/14/2021 08:55'!
log: a

	(a = 0)
		ifTrue: [RSFECError signal: 'zero argument'].
	^ logTable at: (self normalizeIndex: (self maskValue: a)).! !

!RSFECGenericGF methodsFor: 'helper' stamp: 'rww 5/30/2021 08:34'!
maskValue: value

	(value < 0)
 			ifTrue: [^ RSErasureIllegalArgumentError signal: 'Negative argument: ', value].
	^value bitAnd: 16rFF.
! !

!RSFECGenericGF methodsFor: 'arithmetic' stamp: 'rww 5/27/2021 20:27'!
multiply: a by: b

	(a = 0)
		ifTrue: [^ 0]
		ifFalse: [(b = 0)
				ifTrue: [^0]].
	^ self exp: ((self log: a) + (self log: b) \\ (size - 1)).
! !

!RSFECGenericGF methodsFor: 'helper' stamp: 'rww 5/27/2021 08:59'!
normalizeIndex: index

	^ index + 1.! !

!RSFECGenericGF methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:50'!
one

	^ one! !

!RSFECGenericGF methodsFor: 'printing' stamp: 'rww 5/22/2021 07:56'!
printOn: stream

	stream nextPutAll: ( 'GF(', primitive hex, ', ', size, ')').! !

!RSFECGenericGF methodsFor: 'accessing' stamp: 'rww 5/17/2021 16:43'!
size

	^ size! !

!RSFECGenericGF methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:50'!
zero

	^ zero! !

Object subclass: #RSFECGenericGFPoly
	instanceVariableNames: 'field coefficients'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!

!RSFECGenericGFPoly class methodsFor: 'instance creation' stamp: 'rww 6/1/2021 11:39'!
newField: f coefficients: c

	^ self newGenericGFPoly
		initializeField: f coefficients: c asByteArray;
		yourself! !

!RSFECGenericGFPoly class methodsFor: 'instance creation' stamp: 'rww 5/30/2021 19:28'!
newGenericGFPoly

	RSFECGenericGFWithPlugin isPluginAvailable ifTrue: [ ^RSFECGenericGFPolyWithPlugin new ].
	^RSFECGenericGFPoly new! !

!RSFECGenericGFPoly methodsFor: 'arithmetic' stamp: 'rww 6/14/2021 08:55'!
addOrSubtractPoly: other

	| largerCoefficients smallerCoefficients lengthDiff sumDiff indexPlus1 |
	(field = other field)
		ifFalse: [RSFECError signal: 'GenericGFPolys do not have same GenericGF field'].
	self isZero
		ifTrue: [^ other].
	other isZero
		ifTrue: [^ self].

	smallerCoefficients := self coefficients.
	largerCoefficients := other coefficients.
	(smallerCoefficients size > largerCoefficients size)
		ifTrue: [ | temp |
			temp := smallerCoefficients.
			smallerCoefficients := largerCoefficients.
			largerCoefficients := temp].

	sumDiff := largerCoefficients species new: (largerCoefficients size) withAll: 0.
	lengthDiff := largerCoefficients size - smallerCoefficients size.
	(lengthDiff = 0)
		ifFalse: [sumDiff replaceFrom: 1 to: (lengthDiff + 1) with: largerCoefficients startingAt: 1].

	lengthDiff to: (largerCoefficients size - 1)
		do: [:index |
			indexPlus1 := index + 1.
			sumDiff at: indexPlus1 put: (field addOrSubtract: (smallerCoefficients at: (indexPlus1 - lengthDiff)) by: (largerCoefficients at: indexPlus1))].

	^ RSFECGenericGFPoly newField: field coefficients: sumDiff
! !

!RSFECGenericGFPoly methodsFor: 'accessing' stamp: 'rww 5/23/2021 09:59'!
coefficient: degree

	^ coefficients at: (coefficients size - degree)
! !

!RSFECGenericGFPoly methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:44'!
coefficients

	^ coefficients
	! !

!RSFECGenericGFPoly methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:45'!
degree

	^ coefficients size - 1! !

!RSFECGenericGFPoly methodsFor: 'arithmetic' stamp: 'rww 6/2/2021 13:06'!
dividePoly: other

	| denominatorLeadingTerm quotient remainder degreeDifference inverseDenominatorLeadingTerm iterationQuotient scale term |
	(field = other field)
		ifFalse: [^ RSFECError signal: 'GenericGFPolys do not have same GenericGF field'].
	(other isZero)
		ifTrue: [^ RSFECError signal: 'divide by 0'].

	quotient := field zero.
	remainder := self.

	denominatorLeadingTerm := other coefficient: (other degree).
	inverseDenominatorLeadingTerm := field inverse: (denominatorLeadingTerm).

	[(remainder degree >= other degree) & remainder isZero not]
		whileTrue: [
			"Transcript cr; show: 'dividePoly: ', remainder."
			degreeDifference := remainder degree - other degree.
			scale := field multiply: (remainder coefficient: (remainder degree)) by: inverseDenominatorLeadingTerm.
			term := other multiplyByMonomialDegree: degreeDifference coefficient: scale.
			iterationQuotient := field buildMonomialDegree: degreeDifference coefficient: scale.
			quotient := quotient addOrSubtractPoly: iterationQuotient.
			remainder := remainder addOrSubtractPoly: term.
		].

	^ Array with: quotient  with: remainder.! !

!RSFECGenericGFPoly methodsFor: 'evaluating' stamp: 'rww 5/27/2021 09:15'!
evaluateAt: a

	| result aNormalized |
	(a = 0)
		ifTrue: [^ self coefficient: 0].

	(a = 1)
		ifTrue: [
			result := 0.
			coefficients do: [:coefficient |
				result := field addOrSubtract: result by: coefficient].
			^ result].

	aNormalized := a bitAnd: 16rFF.
	result := coefficients at: 1.
	2 to: coefficients size
		do: [:index |
			result := field addOrSubtract: (field multiply: aNormalized by: result) by: (coefficients at: index)].
	^ result! !

!RSFECGenericGFPoly methodsFor: 'accessing' stamp: 'rww 5/21/2021 12:22'!
field

	^ field! !

!RSFECGenericGFPoly methodsFor: 'initialize-release' stamp: 'rww 6/1/2021 11:40'!
initializeField: f coefficients: c
	
	| coefficientsLength firstNonZero |
	(c size = 0)
		ifTrue: [RSFECError signal: 'illegal arguments'].

	field := f.
	coefficientsLength := c size.
	((coefficientsLength > 1) & ((c at: 1) = 0))
		ifTrue: [
			firstNonZero := 2.
			[(firstNonZero <= coefficientsLength) &((c at: firstNonZero) = 0)]
				whileTrue: [firstNonZero := firstNonZero + 1].
			(firstNonZero > coefficientsLength)
				ifTrue: [	coefficients := ByteArray with: 0]
				ifFalse: [coefficients := ByteArray new: (coefficientsLength - firstNonZero + 1) withAll: 0.
					coefficients replaceFrom: 1 to: coefficients size with: c startingAt: firstNonZero]]
		ifFalse: [coefficients := c].
! !

!RSFECGenericGFPoly methodsFor: 'testing' stamp: 'rww 5/16/2021 21:45'!
isZero

	^ (coefficients at: 1) == 0! !

!RSFECGenericGFPoly methodsFor: 'arithmetic' stamp: 'rww 6/1/2021 11:39'!
multiplyByMonomialDegree: degree coefficient: coefficient

	| product size |
	(degree < 0)
		ifTrue: [].
	(coefficient == 0)
		ifTrue: [^ field zero].

	size := coefficients size.
	product := ByteArray new: (size + degree) withAll: 0.
	0 to: (size - 1)
		do: [:index | product at: (index + 1) put: (field multiply: (coefficients at: (index + 1)) by: coefficient)].

	^ RSFECGenericGFPoly newField: field coefficients: product.! !

!RSFECGenericGFPoly methodsFor: 'arithmetic' stamp: 'rww 6/14/2021 08:55'!
multiplyPoly: other

	| aCoeff aCoefficients aLength bCoefficients bLength product |
	(field = other field)
		ifFalse: [RSFECError signal: 'GenericGFPolys do not have same GenericGF field'].
	(self isZero)
		ifTrue: [^ field zero]
		ifFalse: [other isZero ifTrue: [^ field zero]].

	aCoefficients := coefficients.
	aLength := aCoefficients size.
	bCoefficients := other coefficients.
	bLength := bCoefficients size.
	product := bCoefficients species new: (aLength + bLength - 1) withAll: 0.

	1 to: aLength
		do: [:aIndex | 
			aCoeff := aCoefficients at: aIndex.
			1 to: bLength
				do: [:bIndex |
					product at: (aIndex + bIndex - 1) put: ((field 
						addOrSubtract: (product at: (aIndex + bIndex - 1))
						by: (field multiply: aCoeff by: (bCoefficients at: bIndex)))) ]].

	^ RSFECGenericGFPoly newField: field coefficients: product.
! !

!RSFECGenericGFPoly methodsFor: 'arithmetic' stamp: 'rww 5/27/2021 09:16'!
multiplyScalar: scalar

	| product size |
	(scalar = 0)
		ifTrue: [^ field zero].
	(scalar = 1)
		ifTrue: [^ self].

	size := coefficients size.
	product := coefficients species new: size withAll: 0.
	1 to: size 
		do: [:index |
			product at: index put: (field multiply: (coefficients at: index) by: scalar)].
	^ RSFECGenericGFPoly newField: field coefficients: product! !

!RSFECGenericGFPoly methodsFor: 'printing' stamp: 'rww 5/27/2021 09:16'!
printOn: stream

	| coefficient localStream sizeLimit |
	sizeLimit := 1000.
	localStream := LimitedWriteStream on: (String new: (100 min: sizeLimit)).
	localStream setLimit: sizeLimit limitBlock: [
		stream nextPutAll: (localStream setToEnd; contents).
		^ localStream contents].
	
	self degree to: 0 by: -1
		do: [:deg |
			coefficient := self coefficient: deg.
			(coefficient = 0)
				ifFalse: [
					(coefficient < 0)
						ifTrue: [localStream nextPutAll: '  - '.
							coefficient := 0 - coefficient]
						ifFalse: [
							(localStream size > 0)
								ifTrue: [localStream nextPutAll: ' + '.]].
					((deg = 0) | (coefficient ~= 1)) 
						ifTrue: [ | alphaPower |
							alphaPower := field log: coefficient.
							(alphaPower = 0)
								ifTrue: [localStream nextPutAll: '1'.]
								ifFalse: [(alphaPower = 1)
									ifTrue: [localStream nextPutAll: 'a'.]
									ifFalse: [
										localStream nextPutAll: 'a^'.
										localStream nextPutAll: alphaPower asString]]].
					(deg ~= 0)
						ifTrue: [(deg = 1)
							ifTrue: [localStream nextPutAll: 'x'.]
							ifFalse: [
								localStream nextPutAll: 'x^'.
								localStream nextPutAll: deg asString.]]]].
		stream nextPutAll: (localStream setToEnd; contents).
		^ localStream contents.
! !

Object subclass: #RSFECGenericGFPolyWithPlugin
	instanceVariableNames: 'field coefficients'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!

!RSFECGenericGFPolyWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/6/2021 07:43'!
addOrSubtractPoly: other

	| result |
	(field = other field)
		ifFalse: [RSFECError signal: 'GenericGFPolys do not have same GenericGF field'].
	self isZero
		ifTrue: [^ other].
	other isZero
		ifTrue: [^ self].

	result := self
		primAddOrSubtractPolySelfCoefficients: self coefficients 
		otherCoefficients: other coefficients.

	^ RSFECGenericGFPoly newField: field coefficients: result
! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/16/2021 04:05'!
callPrimDvidePoly: other

	| result |
	(field = other field)
		ifFalse: [ ^ RSFECError signal: 'GenericGFPolys do not have same GenericGF field' ].
	(other isZero)
		ifTrue: [ ^ RSFECError signal: 'divide by 0' ].

	[result := self
		primDividePolySelfCoefficients: self coefficients 
		otherCoefficients: other coefficients 
		fieldSize: field size
		zero: field zero coefficients]
			on: Exception
			do: [:ex | ^ RSFECError signal: 'primitive failure: ', ex description ].
	result at: 1 put: (RSFECGenericGFPoly newField: field coefficients: (result at: 1)).
	result at: 2 put: (RSFECGenericGFPoly newField: field coefficients: (result at: 2)).
	^ result! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'accessing' stamp: 'rww 5/23/2021 09:59'!
coefficient: degree

	^ coefficients at: (coefficients size - degree)
! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:44'!
coefficients

	^ coefficients
	! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:45'!
degree

	^ coefficients size - 1! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/1/2021 08:55'!
dividePoly: other

	| denominatorLeadingTerm quotient remainder degreeDifference inverseDenominatorLeadingTerm iterationQuotient scale term |
	(field = other field)
		ifFalse: [^ RSFECError signal: 'GenericGFPolys do not have same GenericGF field'].
	(other isZero)
		ifTrue: [^ RSFECError signal: 'divide by 0'].

	quotient := field zero.
	remainder := self.

	denominatorLeadingTerm := other coefficient: (other degree).
	inverseDenominatorLeadingTerm := field inverse: (denominatorLeadingTerm).

	[(remainder degree >= other degree) & remainder isZero not]
		whileTrue: [
			"Transcript cr; show: 'dividePoly: ', remainder."
			degreeDifference := remainder degree - other degree.
			scale := field multiply: (remainder coefficient: (remainder degree)) by: inverseDenominatorLeadingTerm.
			term := other multiplyByMonomialDegree: degreeDifference coefficient: scale.
			iterationQuotient := field buildMonomialDegree: degreeDifference coefficient: scale.
			quotient := quotient addOrSubtractPoly: iterationQuotient.
			remainder := remainder addOrSubtractPoly: term.
		].

	^ Array with: quotient  with: remainder.! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'evaluating' stamp: 'rww 5/31/2021 14:46'!
evaluateAt: a

	(a = 0)
		ifTrue: [^ self coefficient: 0].
	^ self primEvaluateAt: a coefficients: self coefficients fieldSize: field size. 
! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'accessing' stamp: 'rww 5/21/2021 12:22'!
field

	^ field! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'initialize-release' stamp: 'rww 6/14/2021 09:10'!
initializeField: f coefficients: c
	
	(c size = 0)
		ifTrue: [RSFECError signal: 'illegal arguments'].
	self etrace: #RSFEC msg: 'primitive initialize'.
	field := f.
	coefficients := self primInitializePolyFieldSize: f size coefficients: c.
	! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'testing' stamp: 'rww 5/16/2021 21:45'!
isZero

	^ (coefficients at: 1) == 0! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/6/2021 08:55'!
multiplyByMonomialDegree: degree coefficient: coefficient

	| result |
	(degree < 0)
		ifTrue: [RSFECError signal: 'degree is less than zero'].
	(coefficient == 0)
		ifTrue: [^ field zero].

	result := self 
		primMultiplyByMonomialDegree: degree 
		coefficient: coefficient 
		coefficients: coefficients 
		fieldSize: field size.

	^ RSFECGenericGFPoly newField: field coefficients: result.
! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/6/2021 08:59'!
multiplyPoly: other

	| result |
	(field = other field)
		ifFalse: [RSFECError signal: 'GenericGFPolys do not have same GenericGF field'].
	(self isZero or: [other isZero])
		ifTrue: [^ field zero].

	result := self 
		primMultiplyPolySelfCoefficients: self coefficients
		otherCoefficients: other coefficients
		fieldSize: field size.

	^ RSFECGenericGFPoly newField: field coefficients: result.
! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/6/2021 09:03'!
multiplyScalar: scalar

	| result |
	(scalar = 0)
		ifTrue: [^ field zero].
	(scalar = 1)
		ifTrue: [^ self].

	result := self 
		primMultiplyScalar: scalar 
		coefficients: self coefficients 
		fieldSize: field size.

	^ RSFECGenericGFPoly newField: field coefficients: result.
! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 13:55'!
primAddOrSubtractPolySelfCoefficients: selfCoefficients otherCoefficients: otherCoefficients

	<primitive: 'primitiveFECAddOrSubtractPolySelfCoefficientsOopOtherCoefficientsOop' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 13:56'!
primEvaluateAt: a coefficients: coefficients fieldSize: fieldSize

	<primitive: 'primitiveFECEvaluateAtCoefficientsOopFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 13:57'!
primInitializePolyFieldSize:  fieldSize coefficients: localCoefficients

	<primitive: 'primitiveFECInitializePolyFieldSizeCoefficientsOop' module: 'RSPlugin'>
	^ self primitiveFailed
! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 13:58'!
primMultiplyByMonomialDegree: degree coefficient: coefficient coefficients: coefficients fieldSize: fieldSize

	<primitive: 'primitiveFECMultiplyByMonomialDegreeCoefficientCoefficientsOopFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 13:58'!
primMultiplyPolySelfCoefficients: coefficients otherCoefficients: otherCoefficients fieldSize: fieldSize

	<primitive: 'primitiveFECMultiplyPolySelfCoefficientsOopOtherCoefficientsOopFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 13:59'!
primMultiplyScalar: scalar coefficients: coefficients fieldSize: fieldSize

	<primitive: 'primitiveFECMultiplyScalarCoefficientsOopFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFPolyWithPlugin methodsFor: 'printing' stamp: 'rww 5/27/2021 09:16'!
printOn: stream

	| coefficient localStream sizeLimit |
	sizeLimit := 1000.
	localStream := LimitedWriteStream on: (String new: (100 min: sizeLimit)).
	localStream setLimit: sizeLimit limitBlock: [
		stream nextPutAll: (localStream setToEnd; contents).
		^ localStream contents].
	
	self degree to: 0 by: -1
		do: [:deg |
			coefficient := self coefficient: deg.
			(coefficient = 0)
				ifFalse: [
					(coefficient < 0)
						ifTrue: [localStream nextPutAll: '  - '.
							coefficient := 0 - coefficient]
						ifFalse: [
							(localStream size > 0)
								ifTrue: [localStream nextPutAll: ' + '.]].
					((deg = 0) | (coefficient ~= 1)) 
						ifTrue: [ | alphaPower |
							alphaPower := field log: coefficient.
							(alphaPower = 0)
								ifTrue: [localStream nextPutAll: '1'.]
								ifFalse: [(alphaPower = 1)
									ifTrue: [localStream nextPutAll: 'a'.]
									ifFalse: [
										localStream nextPutAll: 'a^'.
										localStream nextPutAll: alphaPower asString]]].
					(deg ~= 0)
						ifTrue: [(deg = 1)
							ifTrue: [localStream nextPutAll: 'x'.]
							ifFalse: [
								localStream nextPutAll: 'x^'.
								localStream nextPutAll: deg asString.]]]].
		stream nextPutAll: (localStream setToEnd; contents).
		^ localStream contents.
! !

Object subclass: #RSFECGenericGFWithPlugin
	instanceVariableNames: 'zero one size primitive generatorBase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!

!RSFECGenericGFWithPlugin class methodsFor: 'testing' stamp: 'rww 6/5/2021 06:03'!
isPluginAvailable
	"Answer true if this platform has primitive support for the RSFEC GF."
	"RSFECGenericGFWithPlugin isPluginAvailable"

	<primitive: 'primitiveIsPluginAvailable' module: 'RSPlugin'>
	^ false! !

!RSFECGenericGFWithPlugin methodsFor: 'arithmetic' stamp: 'rww 5/30/2021 15:21'!
addOrSubtract: a by: b

	^ self primAddOrSubtract: (self maskValue: a) by: (self maskValue: b).! !

!RSFECGenericGFWithPlugin methodsFor: 'building' stamp: 'rww 6/14/2021 08:55'!
buildMonomialDegree: degree coefficient: coefficient

	| coefficients |
	(degree < 0)
		ifTrue: [RSFECError signal: 'negative degree'].
	(coefficient = zero)
		ifTrue: [^ zero].
	coefficients := ByteArray new: (degree + 1) withAll: 0.
	coefficients at: 1 put: coefficient.
	^ RSFECGenericGFPoly newField: self coefficients: coefficients.! !

!RSFECGenericGFWithPlugin methodsFor: 'arithmetic' stamp: 'rww 5/30/2021 15:25'!
exp: a

	^ self primExp: (self normalizeIndex: a)! !

!RSFECGenericGFWithPlugin methodsFor: 'accessing' stamp: 'rww 5/17/2021 16:43'!
generatorBase

	^ generatorBase! !

!RSFECGenericGFWithPlugin methodsFor: 'initialize-release' stamp: 'rww 5/29/2021 14:01'!
initializePrimitive: prim size: mySize generatorBase: b

	primitive := prim.
	size := mySize.
	generatorBase := b.
	
 	zero := RSFECGenericGFPoly newField: self coefficients: (#( 0) asArray).
	one := RSFECGenericGFPoly newField: self coefficients: (#( 1) asArray).
! !

!RSFECGenericGFWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/5/2021 13:49'!
inverse: a

	^ self primInverse: (self maskValue: a) fieldSize: size! !

!RSFECGenericGFWithPlugin methodsFor: 'arithmetic' stamp: 'rww 5/29/2021 13:01'!
log: a

	^ self primLog: (self maskValue: a)! !

!RSFECGenericGFWithPlugin methodsFor: 'helper' stamp: 'rww 6/14/2021 08:56'!
maskValue: value

	(value < 0)
 			ifTrue: [^ RSFECError signal: 'Negative argument: ', value].
	^value bitAnd: 16rFF.
! !

!RSFECGenericGFWithPlugin methodsFor: 'arithmetic' stamp: 'rww 6/5/2021 13:52'!
multiply: a by: b

	^ self primMultiply: (self maskValue: a) by: (self maskValue: b) fieldSize: size! !

!RSFECGenericGFWithPlugin methodsFor: 'helper' stamp: 'rww 5/27/2021 08:59'!
normalizeIndex: index

	^ index! !

!RSFECGenericGFWithPlugin methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:50'!
one

	^ one! !

!RSFECGenericGFWithPlugin methodsFor: 'primitives' stamp: 'rww 6/5/2021 05:51'!
primAddOrSubtract: a by: b

	<primitive: 'primitiveFECAddOrSubtractBy' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFWithPlugin methodsFor: 'primitives' stamp: 'rww 6/13/2021 13:51'!
primBuildMonomialDegree: degree coefficient: coefficient zero: zeroOop

	<primitive: 'primitiveFECBuildMonomialDegreeCoefficientZeroOop' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFWithPlugin methodsFor: 'primitives' stamp: 'rww 6/5/2021 05:51'!
primExp: a

	<primitive: 'primitiveFECExp' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFWithPlugin methodsFor: 'primitives' stamp: 'rww 6/5/2021 13:49'!
primInverse: a fieldSize: fieldSize

	<primitive: 'primitiveFECInverseFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFWithPlugin methodsFor: 'primitives' stamp: 'rww 6/5/2021 05:52'!
primLog: a

	<primitive: 'primitiveFECLog' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFWithPlugin methodsFor: 'primitives' stamp: 'rww 6/5/2021 13:52'!
primMultiply: a by: b  fieldSize: fieldSize


	<primitive: 'primitiveFECMultiplyByFieldSize' module: 'RSPlugin'>
	^ self primitiveFailed! !

!RSFECGenericGFWithPlugin methodsFor: 'printing' stamp: 'rww 5/27/2021 09:20'!
printOn: stream

	stream nextPutAll: ( 'GF(', primitive hex, ', ', size, ')').! !

!RSFECGenericGFWithPlugin methodsFor: 'accessing' stamp: 'rww 5/17/2021 16:43'!
size

	^ size! !

!RSFECGenericGFWithPlugin methodsFor: 'accessing' stamp: 'rww 5/16/2021 21:50'!
zero

	^ zero! !

Error subclass: #RSFECError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CryptographyRSFEC'!
RSFECGenericGF initialize!
